@@ -1,6 +1,6 @@
 /* print_command -- A way to make readable commands from a command tree. */
 
-/* Copyright (C) 1989-2017 Free Software Foundation, Inc.
+/* Copyright (C) 1989-2022 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -50,58 +50,58 @@
 #include "builtins/common.h"
 
 #if !HAVE_DECL_PRINTF
-extern int printf __P((const char *, ...));	/* Yuck.  Double yuck. */
+extern int printf PARAMS((const char *, ...));	/* Yuck.  Double yuck. */
 #endif
 
 static int indentation;
 static int indentation_amount = 4;
 
 #if defined (PREFER_STDARG)
-typedef void PFUNC __P((const char *, ...));
+typedef void PFUNC PARAMS((const char *, ...));
 
-static void cprintf __P((const char *, ...))  __attribute__((__format__ (printf, 1, 2)));
-static void xprintf __P((const char *, ...))  __attribute__((__format__ (printf, 1, 2)));
+static void cprintf PARAMS((const char *, ...))  __attribute__((__format__ (printf, 1, 2)));
+static void xprintf PARAMS((const char *, ...))  __attribute__((__format__ (printf, 1, 2)));
 #else
 #define PFUNC VFunction
 static void cprintf ();
 static void xprintf ();
 #endif
 
-static void reset_locals __P((void));
-static void newline __P((char *));
-static void indent __P((int));
-static void semicolon __P((void));
-static void the_printed_command_resize __P((int));
+static void reset_locals PARAMS((void));
+static void newline PARAMS((char *));
+static void indent PARAMS((int));
+static void semicolon PARAMS((void));
+static void the_printed_command_resize PARAMS((int));
 
-static void make_command_string_internal __P((COMMAND *));
-static void _print_word_list __P((WORD_LIST *, char *, PFUNC *));
-static void command_print_word_list __P((WORD_LIST *, char *));
-static void print_case_clauses __P((PATTERN_LIST *));
-static void print_redirection_list __P((REDIRECT *));
-static void print_redirection __P((REDIRECT *));
-static void print_heredoc_header __P((REDIRECT *));
-static void print_heredoc_body __P((REDIRECT *));
-static void print_heredocs __P((REDIRECT *));
-static void print_heredoc_bodies __P((REDIRECT *));
-static void print_deferred_heredocs __P((const char *));
+static void make_command_string_internal PARAMS((COMMAND *));
+static void _print_word_list PARAMS((WORD_LIST *, char *, PFUNC *));
+static void command_print_word_list PARAMS((WORD_LIST *, char *));
+static void print_case_clauses PARAMS((PATTERN_LIST *));
+static void print_redirection_list PARAMS((REDIRECT *));
+static void print_redirection PARAMS((REDIRECT *));
+static void print_heredoc_header PARAMS((REDIRECT *));
+static void print_heredoc_body PARAMS((REDIRECT *));
+static void print_heredocs PARAMS((REDIRECT *));
+static void print_heredoc_bodies PARAMS((REDIRECT *));
+static void print_deferred_heredocs PARAMS((const char *));
 
-static void print_for_command __P((FOR_COM *));
+static void print_for_command PARAMS((FOR_COM *));
 #if defined (ARITH_FOR_COMMAND)
-static void print_arith_for_command __P((ARITH_FOR_COM *));
+static void print_arith_for_command PARAMS((ARITH_FOR_COM *));
 #endif
 #if defined (SELECT_COMMAND)
-static void print_select_command __P((SELECT_COM *));
+static void print_select_command PARAMS((SELECT_COM *));
 #endif
-static void print_group_command __P((GROUP_COM *));
-static void print_case_command __P((CASE_COM *));
-static void print_while_command __P((WHILE_COM *));
-static void print_until_command __P((WHILE_COM *));
-static void print_until_or_while __P((WHILE_COM *, char *));
-static void print_if_command __P((IF_COM *));
+static void print_group_command PARAMS((GROUP_COM *));
+static void print_case_command PARAMS((CASE_COM *));
+static void print_while_command PARAMS((WHILE_COM *));
+static void print_until_command PARAMS((WHILE_COM *));
+static void print_until_or_while PARAMS((WHILE_COM *, char *));
+static void print_if_command PARAMS((IF_COM *));
 #if defined (COND_COMMAND)
-static void print_cond_node __P((COND_COM *));
+static void print_cond_node PARAMS((COND_COM *));
 #endif
-static void print_function_def __P((FUNCTION_DEF *));
+static void print_function_def PARAMS((FUNCTION_DEF *));
 
 #define PRINTED_COMMAND_INITIAL_SIZE 64
 #define PRINTED_COMMAND_GROW_SIZE 128
@@ -129,6 +129,7 @@ static int inside_function_def;
 static int skip_this_indent;
 static int was_heredoc;
 static int printing_connection;
+static int printing_comsub;
 static REDIRECT *deferred_heredocs;
 
 /* The depth of the group commands that we are currently printing.  This
@@ -162,6 +163,21 @@ make_command_string (command)
   return (the_printed_command);
 }
 
+/* Print a command substitution after parsing it in parse_comsub to turn it
+   back into an external representation without turning newlines into `;'.
+   Placeholder for other changes, if any are necessary. */
+char *
+print_comsub (command)
+     COMMAND *command;
+{
+  char *ret;
+
+  printing_comsub++;
+  ret = make_command_string (command);
+  printing_comsub--;
+  return ret;
+}
+
 /* The internal function.  This is the real workhorse. */
 static void
 make_command_string_internal (command)
@@ -278,25 +294,39 @@ make_command_string_internal (command)
 	      break;
 
 	    case ';':
-	      if (deferred_heredocs == 0)
-		{
-		  if (was_heredoc == 0)
-		    cprintf (";");
-		  else
-		    was_heredoc = 0;
-		}
-	      else
-		print_deferred_heredocs (inside_function_def ? "" : ";");
+	    case '\n':				/* special case this */
+	      {
+		char c = command->value.Connection->connector;
+		int was_newline;
 
-	      if (inside_function_def)
-		cprintf ("\n");
-	      else
-		{
-		  cprintf (" ");
-		  if (command->value.Connection->second)
-		    skip_this_indent++;
-		}
-	      break;
+		s[0] = printing_comsub ? c : ';';
+		s[1] = '\0';
+
+		was_newline = deferred_heredocs == 0 && was_heredoc == 0 && c == '\n';
+		if (deferred_heredocs == 0)
+		  {
+		    if (was_heredoc == 0)
+		      cprintf ("%s", s);	/* inside_function_def? */
+		    else
+		      was_heredoc = 0;
+		  }
+		else
+		  /* print_deferred_heredocs special-cases `;' */
+		  print_deferred_heredocs (inside_function_def ? "" : ";");
+
+		if (inside_function_def)
+		  cprintf ("\n");
+		else if (printing_comsub && c == '\n' && was_newline == 0)
+		  cprintf ("\n");	/* preserve newlines in comsubs but don't double them */
+		else
+		  {
+		    if (c == ';')
+		      cprintf (" ");
+		    if (command->value.Connection->second)
+		      skip_this_indent++;
+		  }
+		break;
+	      }
 
 	    default:
 	      cprintf (_("print_command: bad connector `%d'"),
@@ -326,7 +356,9 @@ make_command_string_internal (command)
 	  break;
 
 	case cm_coproc:
-	  cprintf ("coproc %s ", command->value.Coproc->name);
+	  cprintf ("coproc ");
+	  if (command->value.Coproc->command->type != cm_simple)
+	    cprintf ("%s ", command->value.Coproc->name);
 	  skip_this_indent++;
 	  make_command_string_internal (command->value.Coproc->command);
 	  break;
@@ -440,7 +472,10 @@ indirection_level_string ()
     change_flag ('x', FLAG_ON);
 
   if (ps4 == 0 || *ps4 == '\0')
-    return (indirection_string);
+    {
+      FREE (ps4);
+      return (indirection_string);
+    }
 
 #if defined (HANDLE_MULTIBYTE)
   ps4_len = strnlen (ps4, MB_CUR_MAX);
@@ -956,11 +991,13 @@ void
 print_simple_command (simple_command)
      SIMPLE_COM *simple_command;
 {
-  command_print_word_list (simple_command->words, " ");
+  if (simple_command->words)
+    command_print_word_list (simple_command->words, " ");
 
   if (simple_command->redirects)
     {
-      cprintf (" ");
+      if (simple_command->words)
+	cprintf (" ");
       print_redirection_list (simple_command->redirects);
     }
 }
@@ -1050,6 +1087,9 @@ print_redirection_list (redirects)
 	  else
 	    hdtail = heredocs = newredir;
 	}
+#if 0
+      /* Remove this heuristic now that the command printing code doesn't
+	 unconditionally put in the redirector file descriptor. */
       else if (redirects->instruction == r_duplicating_output_word && (redirects->flags & REDIR_VARASSIGN) == 0 && redirects->redirector.dest == 1)
 	{
 	  /* Temporarily translate it as the execution code does. */
@@ -1059,6 +1099,7 @@ print_redirection_list (redirects)
 	  print_redirection (redirects);
 	  redirects->instruction = r_duplicating_output_word;
 	}
+#endif
       else
 	print_redirection (redirects);
 
@@ -1217,6 +1258,8 @@ print_redirection (redirect)
     case r_duplicating_input_word:
       if (redirect->rflags & REDIR_VARASSIGN)
 	cprintf ("{%s}<&%s", redir_word->word, redirectee->word);
+      else if (redirector == 0)
+	cprintf ("<&%s", redirectee->word);
       else
 	cprintf ("%d<&%s", redirector, redirectee->word);
       break;
@@ -1224,6 +1267,8 @@ print_redirection (redirect)
     case r_duplicating_output_word:
       if (redirect->rflags & REDIR_VARASSIGN)
 	cprintf ("{%s}>&%s", redir_word->word, redirectee->word);
+      else if (redirector == 1)
+	cprintf (">&%s", redirectee->word);
       else
 	cprintf ("%d>&%s", redirector, redirectee->word);
       break;
@@ -1280,6 +1325,7 @@ reset_locals ()
   indentation = 0;
   printing_connection = 0;
   deferred_heredocs = 0;
+  printing_comsub = 0;
 }
 
 static void
@@ -1312,7 +1358,7 @@ print_function_def (func)
   make_command_string_internal (cmdcopy->type == cm_group
 					? cmdcopy->value.Group->command
 					: cmdcopy);
-  /* XXX - PRINT_DEFERRED_HEREDOCS (""); ? */
+  PRINT_DEFERRED_HEREDOCS ("");
 
   remove_unwind_protect ();
   indentation -= indentation_amount;
@@ -1325,7 +1371,11 @@ print_function_def (func)
       cmdcopy->redirects = func_redirects;
     }
   else
-    newline ("}");
+    {
+      /* { */
+      newline ("}");
+      was_heredoc = 0;		/* not printing any here-documents now */
+    }
 
   dispose_command (cmdcopy);
 }
@@ -1351,6 +1401,7 @@ named_function_string (name, command, flags)
   old_amount = indentation_amount;
   command_string_index = was_heredoc = 0;
   deferred_heredocs = 0;
+  printing_comsub = 0;
 
   if (name && *name)
     {
@@ -1388,7 +1439,7 @@ named_function_string (name, command, flags)
   make_command_string_internal (cmdcopy->type == cm_group
 					? cmdcopy->value.Group->command
 					: cmdcopy);
-  /* XXX - PRINT_DEFERRED_HEREDOCS (""); ? */
+  PRINT_DEFERRED_HEREDOCS ("");
 
   indentation = old_indent;
   indentation_amount = old_amount;
@@ -1401,7 +1452,10 @@ named_function_string (name, command, flags)
       cmdcopy->redirects = func_redirects;
     }
   else
-    newline ("}");
+    {	/* { */
+      newline ("}");
+      was_heredoc = 0;
+    }
 
   result = the_printed_command;
 
