@@ -1,6 +1,6 @@
 /* test.h -- external interface to the conditional command code. */
 
-/* Copyright (C) 1997-2020 Free Software Foundation, Inc.
+/* Copyright (C) 1997-2021 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -27,11 +27,12 @@
 #define TEST_PATMATCH	0x01
 #define TEST_ARITHEXP	0x02
 #define TEST_LOCALE	0x04
+#define TEST_ARRAYEXP	0x08		/* array subscript expansion */
 
 extern int test_unop PARAMS((char *));
 extern int test_binop PARAMS((char *));
 
-extern int unary_test PARAMS((char *, char *));
+extern int unary_test PARAMS((char *, char *, int));
 extern int binary_test PARAMS((char *, char *, char *, int));
 
 extern int test_command PARAMS((int, char **));
